@@ -1,4 +1,3 @@
-from functools import reduce
 from math import ceil
 from re import match
 
@@ -18,6 +17,8 @@ class AutoConvHandler:
         self._bkup_state_routes, self._list_keyboard = None, None
         self.list_labels = None
 
+    # ---- Next state
+
     def _build_keyboard(self, state):
         """Build Keyboard for callback state"""
         cmd_list = (
@@ -50,7 +51,7 @@ class AutoConvHandler:
         if value not in states and -1 not in states:
             raise ValueError(
                 f"Deafult route not found and value {value} "
-                "doesn't exist as route of {state}."
+                f"doesn't exist as route of {state}."
             )
         if value == "BACK" and states.get(value) is True:
             return self.prev_state
@@ -63,25 +64,12 @@ class AutoConvHandler:
             state = self.conversation.get_state(
                 self.tData.context.user_data.get("state")
             )
-            if state.list and isinstance(data, int):
+            if state.list and isinstance(data, int) and not self.tData.update.message:
                 value = [
-                    b
-                    for b in reduce(lambda x, y: x + y, self._list_keyboard)
-                    if b.callback_data is not None and b.callback_data == data
+                    b for b in sum(self._list_keyboard, []) if b.callback_data == data
                 ][0].text
             else:
-                value = (
-                    d
-                    if (c := state.callback)
-                    and (
-                        d := c[0].get(
-                            data
-                            if not isinstance(data, int) and not str(data).isdigit()
-                            else int(data)
-                        )
-                    )
-                    else data
-                )
+                value = state.callback and state.callback[0].get(data) or data
             if state != self.conversation.end:
                 self.tData.context.user_data.get("data").update({state.name: value})
             new_state = self._next_state(state, data)
@@ -95,60 +83,37 @@ class AutoConvHandler:
             data_context.update({"error": False})
         elif data_context.get("error") is not None:
             data_context.pop("error")
-        if self._bkup_state_routes:
-            back_route = self._bkup_state_routes.pop("BACK", None)
-            default_route = self._bkup_state_routes.pop(-1, None)
-            self.conversation.add_routes(
-                self.prev_state,
-                self._bkup_state_routes,
-                default=default_route,
-                back=back_route,
-            )
-        self._bkup_state_routes = None
+        self._restore_basic_routes(self.prev_state)
         return new_state
 
-    def _wrong_message(self):
+    def _wrong_message(self, text=True):
         """Handler for wrong message"""
         self.tData.update.message.delete()
         state = self.conversation.get_state(self.tData.context.user_data.get("state"))
-        if (
-            (self.tData.update.message.text and state.regex_error_text)
-            or state.handler_error_text
-        ) and not self.tData.context.user_data.get("error"):
-            keyboard = self._build_keyboard(state)
+        self._restore_basic_routes(state)
+        error_text = text and state.regex_error_text or state.handler_error_text
+        if error_text and not self.tData.context.user_data.get("error"):
             self.tData.context.user_data.update({"error": True})
-            msg = (
-                state.msg.replace("@@@", a)
-                if state.action and (a := state.action(self.tData.prepare()))
-                else state.msg
-            )
-            reply_msg = f"{msg}\n\n" + (
-                (self.tData.update.message.text and state.regex_error_text)
-                or state.handler_error_text
-            )
-            kwargs = self.conversation.defaults | state.kwargs
-            self.tData.context.user_data.get("bot-msg").edit_text(
-                self.conversation.default_func(reply_msg),
-                reply_markup=InlineKeyboardMarkup(keyboard),
-                **kwargs,
+            self._send_message(
+                state,
+                self.tData.context.user_data.get("bot-msg").edit_text,
+                post_func=lambda x: f"{x}\n\n{error_text}",
             )
         return self.NEXT
 
+    # ---- Dynamic stuff
+
     def _update_dynamic_list(self, state):
         """Update i and routes backup for dynamic list"""
         data = self.tData.context.user_data
         if (state_l := data.get("list")) :
             i = int(data.get("list_i"))
+            new_i = i
             if state.list_all:
                 curr_list = self.list_labels or state_l
                 elem_list = data.get("data").get(state.name)
                 new_i = curr_list.index(elem_list) if elem_list in curr_list else i
-            elif (
-                not state.list_all
-                and data.get("data").get(state.name) not in state.list_buttons
-            ):
-                new_i = i
-            else:
+            elif data.get("data").get(state.name) in state.list_buttons:
                 new_i = (i - 1, i + 1)[
                     data.get("data").get(state.name) == state.list_buttons[1]
                 ]
@@ -159,7 +124,8 @@ class AutoConvHandler:
 
     def _build_dynamic_list(self, state, keyboard):
         """Build dynamic list for current state"""
-        self._bkup_state_routes = self.conversation.routes.get(state.name)
+        if self.prev_state != self.curr_state:
+            self._bkup_state_routes = self.conversation.routes.get(state.name)
         data = self.tData.context.user_data
         state_l = data.get("list")
         basic_routes = {
@@ -169,9 +135,8 @@ class AutoConvHandler:
         }
         for kl in keyboard:
             for button in kl:
-                if (c := button.callback_data) is not None and isinstance(c, int):
+                if (cd := button.callback_data) is not None and isinstance(cd, int):
                     button.callback_data += len(state_l)
-
         self.list_labels = state.list_labels and state.list_labels(self.tData.prepare())
         list_buttons = [
             [
@@ -215,8 +180,8 @@ class AutoConvHandler:
 
     def _build_dynamic_routes(self, state):
         """Build dynamic routes for current state"""
-        ro, de, ba = state.routes(self.tData.prepare())
-        self.conversation.add_routes(state, ro, de, ba)
+        routes, default, back = state.routes(self.tData.prepare())
+        self.conversation.add_routes(state, routes, default, back)
 
     def _build_dynamic_keyboard(self, state):
         """Build dynamic keyboard for current state"""
@@ -230,7 +195,7 @@ class AutoConvHandler:
         if self.prev_state != self.curr_state and data.get("list"):
             data.pop("list")
             data.pop("list_i")
-            self.list_labels = None
+            self.list_labels, self._bkup_state_routes = None, None
         if state.list:
             self._update_dynamic_list(state)
         if state.action:
@@ -249,6 +214,20 @@ class AutoConvHandler:
         )
         return InlineKeyboardMarkup(keyboard), reply_msg
 
+    # ---- Dev functions
+
+    def _restore_basic_routes(self, state):
+        if not self._bkup_state_routes:
+            return
+        back_route = self._bkup_state_routes.pop("BACK", None)
+        default_route = self._bkup_state_routes.pop(-1, None)
+        self.conversation.add_routes(
+            state,
+            self._bkup_state_routes,
+            default=default_route,
+            back=back_route,
+        )
+
     def _init_context(self, state):
         """Initializate user data in context"""
         if not self.tData.context.user_data:
@@ -256,21 +235,34 @@ class AutoConvHandler:
                 {"prev_state": None, "state": state.name, "error": False, "data": {}}
             )
 
-    def _handle_bad_request(self, exception, update, msg, keyboard, state):
+    def _handle_bad_request(self, exception, update, state):
         """Handle every bad request"""
         self.tData.exception = exception
         if match("Message to edit not found", str(exception)):
-            kwargs = self.conversation.defaults | state.kwargs
-            m = update.message.reply_text(
-                self.conversation.default_func(msg), reply_markup=keyboard, **kwargs
-            )
-            self.tData.context.user_data.update({"bot-msg": m})
-            return
+            return self._send_message(state, update.message.reply_text)
         if not match("Message is not modified", str(exception)):
             if self.conversation.fallback_state:
                 return self.force_state(self.conversation.fallback_state, update)
             raise exception
 
+    def _send_message(self, state, send_func, elements=None, save=True, post_func=str):
+        """Main function to send messages"""
+        msg, keyboard = elements or (None, None)
+        if state.long_task:
+            send_func(state.long_task)
+        if not elements:
+            keyboard, msg = self._build_dynamic_stuff(state)
+        new_msg = send_func(
+            self.conversation.default_func(post_func(msg)),
+            reply_markup=keyboard,
+            **(self.conversation.defaults | state.kwargs),
+        )
+        if save:
+            self.tData.context.user_data.update({"bot-msg": new_msg})
+        return new_msg
+
+    # ---- Public functions
+
     def force_state(self, state, update):
         """Force a state in the conversation"""
         if isinstance(state, str):
@@ -279,20 +271,13 @@ class AutoConvHandler:
             # prepare data
             self._init_context(state)
             state = self._change_state(None, state=state)
-            keyboard, reply_msg = self._build_dynamic_stuff(state)
             # check if the message exists
             old_msg = self.tData.context.user_data.get("bot-msg")
             send_msg = old_msg and old_msg.edit_text or update.message.reply_text
             try:
-                kwargs = self.conversation.defaults | state.kwargs
-                m = send_msg(
-                    text=self.conversation.default_func(reply_msg),
-                    reply_markup=keyboard,
-                    **kwargs,
-                )
-                self.tData.context.user_data.update({"bot-msg": m})
+                self._send_message(state, send_msg)
             except BadRequest as e:
-                self._handle_bad_request(e, update, reply_msg, keyboard, state)
+                self._handle_bad_request(e, update, state)
             return self.NEXT
 
     def restart(self, update, context):
@@ -307,14 +292,14 @@ class AutoConvHandler:
 
     def manage_conversation(self, update, context, delete_first=False):
         """Master function for converastion"""
-        reply_msg, keyboard, state = None, None, None
+        state = None
         try:
             self.prev_state = self.prev_state or self.conversation.get_state(
                 context.user_data.get("prev_state")
             )
             self.tData.update_telegram_data(update, context)
             telegram_id = self.tData.update.effective_chat.id
-            # check authorization
+            # ---- check authorization
             if (
                 self.conversation.users_list is not None
                 and telegram_id not in self.conversation.users_list
@@ -322,22 +307,15 @@ class AutoConvHandler:
                 if (fbs := self.conversation.no_auth_state) :
                     return self.force_state(fbs, update)
                 return self.NEXT
-            # start
+            # ---- start | first message
             if not self.tData.context.user_data:
                 if delete_first:
                     update.message.delete()
                 state = self.conversation.start
                 self._init_context(state)
-                keyboard, reply_msg = self._build_dynamic_stuff(state)
-                kwargs = self.conversation.defaults | state.kwargs
-                msg = self.tData.update.message.reply_text(
-                    self.conversation.default_func(reply_msg),
-                    reply_markup=keyboard,
-                    **kwargs,
-                )
-                self.tData.context.user_data.update({"bot-msg": msg})
+                self._send_message(state, self.tData.update.message.reply_text)
                 return self.NEXT
-            # get data
+            # ---- get data | next messages
             state = self.conversation.get_state(
                 self.tData.context.user_data.get("state")
             )
@@ -345,39 +323,25 @@ class AutoConvHandler:
                 data = self.tData.update.callback_query.data
                 to_reply = self.tData.update.callback_query.edit_message_text
             else:
-                data = (state.handler and state.handler(self.tData.prepare())) or (
-                    state.regex and self.tData.update.message.text
-                )
                 if not state.regex and not state.handler:
                     self.tData.update.message.delete()
                     return self.NEXT
-                if (
-                    (
-                        state.regex
-                        and isinstance(data, str)
-                        and not match(state.regex, data)
-                    )
-                    or (state.regex and not data)
-                    or (state.handler and not data)
-                ):
-                    return self._wrong_message()
+                data = (
+                    state.handler
+                    and state.handler(self.tData.prepare())
+                    or self.tData.update.message.text
+                )
+                if state.regex and not match(state.regex, str(data)) or not data:
+                    return self._wrong_message(state.regex)
                 to_reply = self.tData.context.user_data.get("bot-msg").edit_text
                 self.tData.update.message.delete()
-            # next stage
+            # ---- next stage
             typed_data = state.data_type(data) if data != "BACK" else "BACK"
             state = self._change_state(typed_data)
-            if state.long_task:
-                to_reply(state.long_task)
-            keyboard, reply_msg = self._build_dynamic_stuff(state)
-            kwargs = self.conversation.defaults | state.kwargs
-            to_reply(
-                self.conversation.default_func(reply_msg),
-                reply_markup=keyboard,
-                **kwargs,
-            )
+            self._send_message(state, to_reply, save=False)
             if state == self.conversation.end:
                 self.tData.context.user_data.clear()
                 return ConversationHandler.END
             return self.NEXT
         except (Exception, BadRequest) as e:
-            self._handle_bad_request(e, update, reply_msg, keyboard, state)
+            self._handle_bad_request(e, update, state)
